#ifndef QTENGINE_QT_WRAPPER_H
#define QTENGINE_QT_WRAPPER_H


#include <gtk/gtknotebook.h>
#include <gdk/gdkgc.h>
#include <gtk/gtkstyle.h>

#ifdef __cplusplus
extern "C" {

class CachedImage
{
	// This class helps speed up image drawing.  It holds one
	// pixmap, in a format ready to be passed to GDK, of a widget
	// of a specified size and type

public:
	enum WidgetType
	{
		Button,
		Tab,
		LineEdit,
		ComboBox,
		Frame,
		CheckBox,
		RadioButton,
		MenuItem,
		MenuBar,
		MenuBarItem,
		Splitter,
		TabFrame,
		Menu,
		ProgressBar,
		Slider,
		SpinButton,
		Arrow,
		ListHeader,
		ToolButton,
		MenuCheck
	};

	CachedImage() {};
	CachedImage(WidgetType t, GtkStateType s, int w, int h, QImage i, unsigned flags = 0, float percentage_ = 0);

	WidgetType type;
	GtkStateType state;
	int w;
	int h;
	QImage image;

	// In cache_find, if you care about a specific attrib, ask with
	// flag = attrib | (either test_on or test_off)
	enum {
		is_checked = 1 << 0,
		has_focus  = 1 << 1
	};
	unsigned flags;
	
	float percentage;
};

void cacheAppend(CachedImage image);

void rearrangeColorComponents(QImage &image);
void mapColour(GdkColor* g, QColor q);

#endif

void createQApp();
void destroyQApp();
void setColors(GtkStyle* style);
void setRcProperties(GtkRcStyle* rc_style);
void drawButton(GdkWindow * window, GtkStyle * style, GtkStateType state, int x, int y, int w, int h);
void drawToolButton(GdkWindow * window, GtkStyle * style, GtkStateType state, int x, int y, int w, int h);
void drawTab(GdkWindow * window, GtkStyle * style, GtkStateType state, int x, int y, int w, int h);
void drawVLine(GdkWindow * window, GtkStyle * style, GtkStateType state, int x, int ySource, int yDest);
void drawHLine(GdkWindow * window, GtkStyle * style, GtkStateType state, int y, int xSource, int xDest);
void drawLineEdit(GdkWindow * window, GtkStyle * style, GtkStateType state, int hasFocus, int x, int y, int w, int h);
void drawComboBox(GdkWindow * window, GtkStyle * style, GtkStateType state, int x, int y, int w, int h);
void drawFrame(GdkWindow * window, GtkStyle * style, GtkStateType state, GtkShadowType shadow_type, int x, int y, int w, int h);
void drawCheckBox(GdkWindow * window, GtkStyle * style, GtkStateType state, int on, int x, int y, int w, int h);
void drawMenuCheck(GdkWindow * window, GtkStyle * style, GtkStateType state, int x, int y, int w, int h);
void drawRadioButton(GdkWindow * window, GtkStyle * style, GtkStateType state, int on, int x, int y, int w, int h);
void drawScrollBar(GdkWindow * window, GtkStyle * style, GtkStateType state, int orientation, GtkAdjustment* adj, int x, int y, int w, int h);
void drawSplitter(GdkWindow * window, GtkStyle * style, GtkStateType state, int orientation, int x, int y, int w, int h);
void drawMenuBarItem(GdkWindow * window, GtkStyle * style, GtkStateType state, int x, int y, int w, int h);
void drawMenuItem(GdkWindow * window, GtkStyle * style, GtkStateType state, int x, int y, int w, int h);
void drawMenu(GdkWindow * window, GtkStyle * style, GtkStateType state, int x, int y, int w, int h);
void drawTabFrame(GdkWindow * window, GtkStyle * style, GtkStateType state, int x, int y, int w, int h);
void drawProgressBar(GdkWindow * window, GtkStyle * style, GtkStateType state, gfloat percentage, int x, int y, int w, int h);
void drawSlider(GdkWindow * window, GtkStyle * style, GtkStateType state, GtkAdjustment* adj, int x, int y, int w, int h);
void drawSpinButton(GdkWindow * window, GtkStyle * style, GtkStateType state, int direction, int x, int y, int w, int h);
void drawArrow(GdkWindow * window, GtkStyle * style, GtkStateType state, GtkArrowType direction, int x, int y, int w, int h);
void drawListHeader(GdkWindow * window, GtkStyle * style, GtkStateType state, int x, int y, int w, int h);

GdkGC* alternateBackgroundGc(GtkStyle* style);

#ifdef  __cplusplus
}
#endif

#endif
